@@ -1,15 +1,19 @@
+# -*- coding: utf-8 -*-
 # Copyright (c) 2009-2011, 2013-2014 LOGILAB S.A. (Paris, FRANCE) <contact@logilab.fr>
 # Copyright (c) 2012 FELD Boris <lothiraldan@gmail.com>
-# Copyright (c) 2014-2018 Claudiu Popa <pcmanticore@gmail.com>
+# Copyright (c) 2014-2020 Claudiu Popa <pcmanticore@gmail.com>
 # Copyright (c) 2014 Google, Inc.
 # Copyright (c) 2014 Eevee (Alex Munroe) <amunroe@yelp.com>
 # Copyright (c) 2015-2016 Ceridwen <ceridwenv@gmail.com>
 # Copyright (c) 2015 Florian Bruhin <me@the-compiler.org>
 # Copyright (c) 2016-2017 Derek Gustafson <degustaf@gmail.com>
 # Copyright (c) 2017 Calen Pennington <calen.pennington@gmail.com>
+# Copyright (c) 2018-2019 hippo91 <guillaume.peillex@gmail.com>
+# Copyright (c) 2018 Ville Skyttä <ville.skytta@iki.fi>
 # Copyright (c) 2018 Bryce Guinta <bryce.paul.guinta@gmail.com>
 # Copyright (c) 2018 Nick Drozd <nicholasdrozd@gmail.com>
 # Copyright (c) 2018 Daniel Colascione <dancol@dancol.org>
+# Copyright (c) 2019 Hugo van Kemenade <hugovk@users.noreply.github.com>
 
 # Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
 # For details: https://github.com/PyCQA/astroid/blob/master/COPYING.LESSER
@@ -20,7 +24,6 @@ inference utils.
 
 import builtins
 import collections
-import sys
 
 from astroid import context as contextmod
 from astroid import exceptions
@@ -32,13 +35,10 @@ BUILTINS = builtins.__name__
 manager = util.lazy_import("manager")
 MANAGER = manager.AstroidManager()
 
-if sys.version_info >= (3, 0):
-    # TODO: check if needs special treatment
-    BUILTINS = "builtins"
-    BOOL_SPECIAL_METHOD = "__bool__"
-else:
-    BUILTINS = "__builtin__"
-    BOOL_SPECIAL_METHOD = "__nonzero__"
+# TODO: check if needs special treatment
+BUILTINS = "builtins"
+BOOL_SPECIAL_METHOD = "__bool__"
+
 PROPERTIES = {BUILTINS + ".property", "abc.abstractproperty"}
 # List of possible property names. We use this list in order
 # to see if a method is a property or not. This should be
@@ -62,13 +62,12 @@ POSSIBLE_PROPERTIES = {
 }
 
 
-def _is_property(meth):
-    if PROPERTIES.intersection(meth.decoratornames()):
+def _is_property(meth, context=None):
+    decoratornames = meth.decoratornames(context=context)
+    if PROPERTIES.intersection(decoratornames):
         return True
     stripped = {
-        name.split(".")[-1]
-        for name in meth.decoratornames()
-        if name is not util.Uninferable
+        name.split(".")[-1] for name in decoratornames if name is not util.Uninferable
     }
     if any(name in stripped for name in POSSIBLE_PROPERTIES):
         return True
@@ -77,7 +76,7 @@ def _is_property(meth):
     if not meth.decorators:
         return False
     for decorator in meth.decorators.nodes or ():
-        inferred = helpers.safe_infer(decorator)
+        inferred = helpers.safe_infer(decorator, context=context)
         if inferred is None or inferred is util.Uninferable:
             continue
         if inferred.__class__.__name__ == "ClassDef":
@@ -211,7 +210,11 @@ class BaseInstance(Proxy):
         try:
             # avoid recursively inferring the same attr on the same class
             if context.push((self._proxied, name)):
-                return
+                raise exceptions.InferenceError(
+                    message="Cannot infer the same attribute again",
+                    node=self,
+                    context=context,
+                )
 
             # XXX frame should be self._proxied, or not ?
             get_attr = self.getattr(name, context, lookupclass=False)
@@ -224,7 +227,7 @@ class BaseInstance(Proxy):
                 # descriptors
                 # But only if the _proxied is the Class.
                 if self._proxied.__class__.__name__ != "ClassDef":
-                    raise exceptions.InferenceError(**vars(error)) from error
+                    raise
                 attrs = self._proxied.igetattr(name, context, class_context=False)
                 yield from self._wrap_attr(attrs, context)
             except exceptions.AttributeInferenceError as error:
@@ -239,15 +242,9 @@ class BaseInstance(Proxy):
                 else:
                     yield BoundMethod(attr, self)
             elif hasattr(attr, "name") and attr.name == "<lambda>":
-                # This is a lambda function defined at class level,
-                # since its scope is the underlying _proxied class.
-                # Unfortunately, we can't do an isinstance check here,
-                # because of the circular dependency between astroid.bases
-                # and astroid.scoped_nodes.
-                if attr.statement().scope() == self._proxied:
-                    if attr.args.args and attr.args.args[0].name == "self":
-                        yield BoundMethod(attr, self)
-                        continue
+                if attr.args.arguments and attr.args.arguments[0].name == "self":
+                    yield BoundMethod(attr, self)
+                    continue
                 yield attr
             else:
                 yield attr
@@ -295,7 +292,7 @@ class Instance(BaseInstance):
     def display_type(self):
         return "Instance of"
 
-    def bool_value(self):
+    def bool_value(self, context=None):
         """Infer the truth value for an Instance
 
         The truth value of an instance is determined by these conditions:
@@ -308,7 +305,7 @@ class Instance(BaseInstance):
              nonzero. If a class defines neither __len__() nor __bool__(),
              all its instances are considered true.
         """
-        context = contextmod.InferenceContext()
+        context = context or contextmod.InferenceContext()
         context.callcontext = contextmod.CallContext(args=[])
         context.boundnode = self
 
@@ -383,7 +380,7 @@ class UnboundMethod(Proxy):
             return (Instance(x) if x is not util.Uninferable else x for x in infer)
         return self._proxied.infer_call_result(caller, context)
 
-    def bool_value(self):
+    def bool_value(self, context=None):
         return True
 
 
@@ -398,6 +395,9 @@ class BoundMethod(UnboundMethod):
         self.bound = bound
 
     def implicit_parameters(self):
+        if self.name == "__new__":
+            # __new__ acts as a classmethod but the class argument is not implicit.
+            return 0
         return 1
 
     def is_bound(self):
@@ -410,6 +410,7 @@ class BoundMethod(UnboundMethod):
         a subtype of ``type``, the name needs to be a string, the bases
         needs to be a tuple of classes
         """
+        # pylint: disable=import-outside-toplevel; circular import
         from astroid import node_classes
 
         # Verify the metaclass
@@ -487,7 +488,7 @@ class BoundMethod(UnboundMethod):
 
         return super(BoundMethod, self).infer_call_result(caller, context)
 
-    def bool_value(self):
+    def bool_value(self, context=None):
         return True
 
 
@@ -513,7 +514,7 @@ class Generator(BaseInstance):
     def display_type(self):
         return "Generator"
 
-    def bool_value(self):
+    def bool_value(self, context=None):
         return True
 
     def __repr__(self):
