@@ -2,7 +2,7 @@
 # Copyright (c) 2006-2014 LOGILAB S.A. (Paris, FRANCE) <contact@logilab.fr>
 # Copyright (c) 2010 Daniel Harding <dharding@gmail.com>
 # Copyright (c) 2011, 2013-2015 Google, Inc.
-# Copyright (c) 2013-2018 Claudiu Popa <pcmanticore@gmail.com>
+# Copyright (c) 2013-2020 Claudiu Popa <pcmanticore@gmail.com>
 # Copyright (c) 2013 Phil Schaf <flying-sheep@web.de>
 # Copyright (c) 2014 Eevee (Alex Munroe) <amunroe@yelp.com>
 # Copyright (c) 2015-2016 Florian Bruhin <me@the-compiler.org>
@@ -14,9 +14,12 @@
 # Copyright (c) 2017-2018 Ashley Whetter <ashley@awhetter.co.uk>
 # Copyright (c) 2017 Łukasz Rogalski <rogalski.91@gmail.com>
 # Copyright (c) 2017 David Euresti <david@dropbox.com>
-# Copyright (c) 2018 Nick Drozd <nicholasdrozd@gmail.com>
+# Copyright (c) 2018-2019 Nick Drozd <nicholasdrozd@gmail.com>
+# Copyright (c) 2018 Ville Skyttä <ville.skytta@iki.fi>
 # Copyright (c) 2018 Anthony Sottile <asottile@umich.edu>
 # Copyright (c) 2018 HoverHell <hoverhell@gmail.com>
+# Copyright (c) 2019 Hugo van Kemenade <hugovk@users.noreply.github.com>
+# Copyright (c) 2019 Peter de Blanc <peter@standard.ai>
 
 # Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
 # For details: https://github.com/PyCQA/astroid/blob/master/COPYING.LESSER
@@ -48,6 +51,11 @@ from astroid import util
 
 BUILTINS = builtins.__name__
 ITER_METHODS = ("__iter__", "__getitem__")
+EXCEPTION_BASE_CLASSES = frozenset({"Exception", "BaseException"})
+objects = util.lazy_import("objects")
+BUILTIN_DESCRIPTORS = frozenset(
+    {"classmethod", "staticmethod", "builtins.classmethod", "builtins.staticmethod"}
+)
 
 
 def _c3_merge(sequences, cls, context):
@@ -88,22 +96,32 @@ def _c3_merge(sequences, cls, context):
     return None
 
 
-def _verify_duplicates_mro(sequences, cls, context):
+def clean_duplicates_mro(sequences, cls, context):
     for sequence in sequences:
-        names = [(node.lineno, node.qname()) for node in sequence if node.name]
-        if len(names) != len(set(names)):
+        names = [
+            (node.lineno, node.qname()) if node.name else None for node in sequence
+        ]
+        last_index = dict(map(reversed, enumerate(names)))
+        if names and names[0] is not None and last_index[names[0]] != 0:
             raise exceptions.DuplicateBasesError(
                 message="Duplicates found in MROs {mros} for {cls!r}.",
                 mros=sequences,
                 cls=cls,
                 context=context,
             )
+        yield [
+            node
+            for i, (node, name) in enumerate(zip(sequence, names))
+            if name is None or last_index[name] == i
+        ]
 
 
 def function_to_method(n, klass):
     if isinstance(n, FunctionDef):
         if n.type == "classmethod":
             return bases.BoundMethod(n, klass)
+        if n.type == "property":
+            return n
         if n.type != "staticmethod":
             return bases.UnboundMethod(n)
     return n
@@ -512,6 +530,11 @@ class Module(LocalsDictNodeNG):
         return "Module"
 
     def getattr(self, name, context=None, ignore_locals=False):
+        if not name:
+            raise exceptions.AttributeInferenceError(
+                target=self, attribute=name, context=context
+            )
+
         result = []
         name_in_locals = name in self.locals
 
@@ -722,7 +745,7 @@ class Module(LocalsDictNodeNG):
         """
         return [name for name in self.keys() if not name.startswith("_")]
 
-    def bool_value(self):
+    def bool_value(self, context=None):
         """Determine the boolean value of this node.
 
         :returns: The boolean value of this node.
@@ -808,7 +831,7 @@ class GeneratorExp(ComprehensionScope):
         else:
             self.generators = generators
 
-    def bool_value(self):
+    def bool_value(self, context=None):
         """Determine the boolean value of this node.
 
         :returns: The boolean value of this node.
@@ -888,7 +911,7 @@ class DictComp(ComprehensionScope):
         else:
             self.generators = generators
 
-    def bool_value(self):
+    def bool_value(self, context=None):
         """Determine the boolean value of this node.
 
         :returns: The boolean value of this node.
@@ -960,7 +983,7 @@ class SetComp(ComprehensionScope):
         else:
             self.generators = generators
 
-    def bool_value(self):
+    def bool_value(self, context=None):
         """Determine the boolean value of this node.
 
         :returns: The boolean value of this node.
@@ -1007,7 +1030,7 @@ class _ListComp(node_classes.NodeNG):
         self.elt = elt
         self.generators = generators
 
-    def bool_value(self):
+    def bool_value(self, context=None):
         """Determine the boolean value of this node.
 
         :returns: The boolean value of this node.
@@ -1061,6 +1084,21 @@ def _infer_decorator_callchain(node):
             return "classmethod"
         if result.is_subtype_of("%s.staticmethod" % BUILTINS):
             return "staticmethod"
+    if isinstance(result, FunctionDef):
+        if not result.decorators:
+            return None
+        # Determine if this function is decorated with one of the builtin descriptors we want.
+        for decorator in result.decorators.nodes:
+            if isinstance(decorator, node_classes.Name):
+                if decorator.name in BUILTIN_DESCRIPTORS:
+                    return decorator.name
+            if (
+                isinstance(decorator, node_classes.Attribute)
+                and isinstance(decorator.expr, node_classes.Name)
+                and decorator.expr.name == BUILTINS
+                and decorator.attrname in BUILTIN_DESCRIPTORS
+            ):
+                return decorator.attrname
     return None
 
 
@@ -1089,7 +1127,7 @@ class Lambda(mixins.FilterStmtsMixin, LocalsDictNodeNG):
         :rtype: str
         """
         # pylint: disable=no-member
-        if self.args.args and self.args.args[0].name == "self":
+        if self.args.arguments and self.args.arguments[0].name == "self":
             if isinstance(self.parent.scope(), ClassDef):
                 return "method"
         return "function"
@@ -1178,8 +1216,8 @@ class Lambda(mixins.FilterStmtsMixin, LocalsDictNodeNG):
         # args is in fact redefined later on by postinit. Can't be changed
         # to None due to a strong interaction between Lambda and FunctionDef.
 
-        if self.args.args:  # maybe None with builtin functions
-            names = _rec_get_names(self.args.args)
+        if self.args.arguments:  # maybe None with builtin functions
+            names = _rec_get_names(self.args.arguments)
         else:
             names = []
         if self.args.vararg:
@@ -1231,7 +1269,7 @@ class Lambda(mixins.FilterStmtsMixin, LocalsDictNodeNG):
             frame = self
         return frame._scope_lookup(node, name, offset)
 
-    def bool_value(self):
+    def bool_value(self, context=None):
         """Determine the boolean value of this node.
 
         :returns: The boolean value of this node.
@@ -1367,9 +1405,6 @@ class FunctionDef(mixins.MultiLineBlockMixin, node_classes.Statement, Lambda):
         self.type_comment_returns = type_comment_returns
         self.type_comment_args = type_comment_args
 
-        if isinstance(self.parent.frame(), ClassDef):
-            self.set_local("__class__", self.parent.frame())
-
     @decorators_mod.cachedproperty
     def extra_decorators(self):
         """The extra decorators that this function can have.
@@ -1414,17 +1449,17 @@ class FunctionDef(mixins.MultiLineBlockMixin, node_classes.Statement, Lambda):
         return decorators
 
     @decorators_mod.cachedproperty
-    def type(self):
+    def type(
+        self
+    ):  # pylint: disable=invalid-overridden-method,too-many-return-statements
         """The function type for this node.
 
         Possible values are: method, function, staticmethod, classmethod.
 
         :type: str
         """
-        builtin_descriptors = {"classmethod", "staticmethod"}
-
         for decorator in self.extra_decorators:
-            if decorator.func.name in builtin_descriptors:
+            if decorator.func.name in BUILTIN_DESCRIPTORS:
                 return decorator.func.name
 
         frame = self.parent.frame()
@@ -1442,8 +1477,15 @@ class FunctionDef(mixins.MultiLineBlockMixin, node_classes.Statement, Lambda):
 
         for node in self.decorators.nodes:
             if isinstance(node, node_classes.Name):
-                if node.name in builtin_descriptors:
+                if node.name in BUILTIN_DESCRIPTORS:
                     return node.name
+            if (
+                isinstance(node, node_classes.Attribute)
+                and isinstance(node.expr, node_classes.Name)
+                and node.expr.name == BUILTINS
+                and node.attrname in BUILTIN_DESCRIPTORS
+            ):
+                return node.attrname
 
             if isinstance(node, node_classes.Call):
                 # Handle the following case:
@@ -1517,10 +1559,18 @@ class FunctionDef(mixins.MultiLineBlockMixin, node_classes.Statement, Lambda):
         """this method doesn't look in the instance_attrs dictionary since it's
         done by an Instance proxy at inference time.
         """
+        if not name:
+            raise exceptions.AttributeInferenceError(
+                target=self, attribute=name, context=context
+            )
+
+        found_attrs = []
         if name in self.instance_attrs:
-            return self.instance_attrs[name]
+            found_attrs = self.instance_attrs[name]
         if name in self.special_attributes:
-            return [self.special_attributes.lookup(name)]
+            found_attrs.append(self.special_attributes.lookup(name))
+        if found_attrs:
+            return found_attrs
         raise exceptions.AttributeInferenceError(target=self, attribute=name)
 
     def igetattr(self, name, context=None):
@@ -1543,9 +1593,11 @@ class FunctionDef(mixins.MultiLineBlockMixin, node_classes.Statement, Lambda):
         return self.type != "function" and isinstance(self.parent.frame(), ClassDef)
 
     @decorators_mod.cached
-    def decoratornames(self):
+    def decoratornames(self, context=None):
         """Get the qualified names of each of the decorators on this function.
 
+        :param context:
+            An inference context that can be passed to inference functions
         :returns: The names of the decorators.
         :rtype: set(str)
         """
@@ -1556,7 +1608,7 @@ class FunctionDef(mixins.MultiLineBlockMixin, node_classes.Statement, Lambda):
         decoratornodes += self.extra_decorators
         for decnode in decoratornodes:
             try:
-                for infnode in decnode.infer():
+                for infnode in decnode.infer(context=context):
                     result.add(infnode.qname())
             except exceptions.InferenceError:
                 continue
@@ -1654,7 +1706,13 @@ class FunctionDef(mixins.MultiLineBlockMixin, node_classes.Statement, Lambda):
 
         first_return = next(returns, None)
         if not first_return:
-            raise exceptions.InferenceError("Empty return iterator")
+            if self.body and isinstance(self.body[-1], node_classes.Assert):
+                yield node_classes.Const(None)
+                return
+
+            raise exceptions.InferenceError(
+                "The function does not have any return statements"
+            )
 
         for returnnode in itertools.chain((first_return,), returns):
             if returnnode.value is None:
@@ -1665,7 +1723,7 @@ class FunctionDef(mixins.MultiLineBlockMixin, node_classes.Statement, Lambda):
                 except exceptions.InferenceError:
                     yield util.Uninferable
 
-    def bool_value(self):
+    def bool_value(self, context=None):
         """Determine the boolean value of this node.
 
         :returns: The boolean value of this node.
@@ -1685,6 +1743,18 @@ class FunctionDef(mixins.MultiLineBlockMixin, node_classes.Statement, Lambda):
 
         yield from self.body
 
+    def scope_lookup(self, node, name, offset=0):
+        """Lookup where the given name is assigned."""
+        if name == "__class__":
+            # __class__ is an implicit closure reference created by the compiler
+            # if any methods in a class body refer to either __class__ or super.
+            # In our case, we want to be able to look it up in the current scope
+            # when `__class__` is being used.
+            frame = self.parent.frame()
+            if isinstance(frame, ClassDef):
+                return self, [frame]
+        return super().scope_lookup(node, name, offset)
+
 
 class AsyncFunctionDef(FunctionDef):
     """Class representing an :class:`ast.FunctionDef` node.
@@ -1730,8 +1800,8 @@ def _is_metaclass(klass, seen=None):
                 baseobj_name = baseobj.qname()
                 if baseobj_name in seen:
                     continue
-                else:
-                    seen.add(baseobj_name)
+
+                seen.add(baseobj_name)
                 if isinstance(baseobj, bases.Instance):
                     # not abstract
                     return False
@@ -1921,9 +1991,8 @@ class ClassDef(mixins.FilterStmtsMixin, LocalsDictNodeNG, node_classes.Statement
         :rtype: tuple(tuple(str, node_classes.Const), ...)
         """
         locals_ = (("__module__", self.special_attributes.attr___module__),)
-        if sys.version_info >= (3, 3):
-            # __qualname__ is defined in PEP3155
-            locals_ += (("__qualname__", self.special_attributes.attr___qualname__),)
+        # __qualname__ is defined in PEP3155
+        locals_ += (("__qualname__", self.special_attributes.attr___qualname__),)
         return locals_
 
     # pylint: disable=redefined-outer-name
@@ -2064,7 +2133,14 @@ class ClassDef(mixins.FilterStmtsMixin, LocalsDictNodeNG, node_classes.Statement
         # Get the bases of the class.
         class_bases = next(caller.args[1].infer(context))
         if isinstance(class_bases, (node_classes.Tuple, node_classes.List)):
-            result.bases = class_bases.itered()
+            bases = []
+            for base in class_bases.itered():
+                inferred = next(base.infer(context=context))
+                if inferred:
+                    bases.append(
+                        node_classes.EvaluatedObject(original=base, value=inferred)
+                    )
+            result.bases = bases
         else:
             # There is currently no AST node that can represent an 'unknown'
             # node (Uninferable is not an AST node), therefore we simply return Uninferable here
@@ -2102,13 +2178,14 @@ class ClassDef(mixins.FilterStmtsMixin, LocalsDictNodeNG, node_classes.Statement
                 dunder_call = next(metaclass.igetattr("__call__", context))
         except exceptions.AttributeInferenceError:
             pass
+
         if dunder_call and dunder_call.qname() != "builtins.type.__call__":
+            # Call type.__call__ if not set metaclass
+            # (since type is the default metaclass)
             context = contextmod.bind_context_to_node(context, self)
             yield from dunder_call.infer_call_result(caller, context)
         else:
-            # Call type.__call__ if not set metaclass
-            # (since type is the default metaclass)
-            yield bases.Instance(self)
+            yield self.instantiate_class()
 
     def scope_lookup(self, node, name, offset=0):
         """Lookup where the given name is assigned.
@@ -2318,6 +2395,12 @@ class ClassDef(mixins.FilterStmtsMixin, LocalsDictNodeNG, node_classes.Statement
             or self if this is not possible.
         :rtype: Instance or ClassDef
         """
+        try:
+            if any(cls.name in EXCEPTION_BASE_CLASSES for cls in self.mro()):
+                # Subclasses of exceptions can be exception instances
+                return objects.ExceptionInstance(self)
+        except exceptions.MroError:
+            pass
         return bases.Instance(self)
 
     def getattr(self, name, context=None, class_context=True):
@@ -2346,6 +2429,11 @@ class ClassDef(mixins.FilterStmtsMixin, LocalsDictNodeNG, node_classes.Statement
 
         :raises AttributeInferenceError: If the attribute cannot be inferred.
         """
+        if not name:
+            raise exceptions.AttributeInferenceError(
+                target=self, attribute=name, context=context
+            )
+
         values = self.locals.get(name, [])
         if name in self.special_attributes and class_context and not values:
             result = [self.special_attributes.lookup(name)]
@@ -2367,13 +2455,23 @@ class ClassDef(mixins.FilterStmtsMixin, LocalsDictNodeNG, node_classes.Statement
             raise exceptions.AttributeInferenceError(
                 target=self, attribute=name, context=context
             )
+
+        # Look for AnnAssigns, which are not attributes in the purest sense.
+        for value in values:
+            if isinstance(value, node_classes.AssignName):
+                stmt = value.statement()
+                if isinstance(stmt, node_classes.AnnAssign) and stmt.value is None:
+                    raise exceptions.AttributeInferenceError(
+                        target=self, attribute=name, context=context
+                    )
         return values
 
     def _metaclass_lookup_attribute(self, name, context):
         """Search the given name in the implicit and the explicit metaclass."""
         attrs = set()
         implicit_meta = self.implicit_metaclass()
-        metaclass = self.metaclass()
+        context = contextmod.copy_context(context)
+        metaclass = self.metaclass(context=context)
         for cls in {implicit_meta, metaclass}:
             if cls and cls != self and isinstance(cls, ClassDef):
                 cls_attributes = self._get_attribute_from_metaclass(cls, name, context)
@@ -2391,8 +2489,8 @@ class ClassDef(mixins.FilterStmtsMixin, LocalsDictNodeNG, node_classes.Statement
                 yield attr
                 continue
 
-            if bases._is_property(attr):
-                yield from attr.infer_call_result(self, context)
+            if isinstance(attr, objects.Property):
+                yield attr
                 continue
             if attr.type == "classmethod":
                 # If the method is a classmethod, then it will
@@ -2420,9 +2518,23 @@ class ClassDef(mixins.FilterStmtsMixin, LocalsDictNodeNG, node_classes.Statement
         # instance
         context = contextmod.copy_context(context)
         context.lookupname = name
+
+        metaclass = self.declared_metaclass(context=context)
         try:
-            attr = self.getattr(name, context, class_context=class_context)[0]
-            for inferred in bases._infer_stmts([attr], context, frame=self):
+            attributes = self.getattr(name, context, class_context=class_context)
+            # If we have more than one attribute, make sure that those starting from
+            # the second one are from the same scope. This is to account for modifications
+            # to the attribute happening *after* the attribute's definition (e.g. AugAssigns on lists)
+            if len(attributes) > 1:
+                first_attr, attributes = attributes[0], attributes[1:]
+                first_scope = first_attr.scope()
+                attributes = [first_attr] + [
+                    attr
+                    for attr in attributes
+                    if attr.parent and attr.parent.scope() == first_scope
+                ]
+
+            for inferred in bases._infer_stmts(attributes, context, frame=self):
                 # yield Uninferable object instead of descriptors when necessary
                 if not isinstance(inferred, node_classes.Const) and isinstance(
                     inferred, bases.Instance
@@ -2433,6 +2545,23 @@ class ClassDef(mixins.FilterStmtsMixin, LocalsDictNodeNG, node_classes.Statement
                         yield inferred
                     else:
                         yield util.Uninferable
+                elif isinstance(inferred, objects.Property):
+                    function = inferred.function
+                    if not class_context:
+                        # Through an instance so we can solve the property
+                        yield from function.infer_call_result(
+                            caller=self, context=context
+                        )
+                    # If we have a metaclass, we're accessing this attribute through
+                    # the class itself, which means we can solve the property
+                    elif metaclass:
+                        # Resolve a property as long as it is not accessed through
+                        # the class itself.
+                        yield from function.infer_call_result(
+                            caller=self, context=context
+                        )
+                    else:
+                        yield inferred
                 else:
                     yield function_to_method(inferred, self)
         except exceptions.AttributeInferenceError as error:
@@ -2705,7 +2834,7 @@ class ClassDef(mixins.FilterStmtsMixin, LocalsDictNodeNG, node_classes.Statement
         if not all(slot is not None for slot in slots):
             return None
 
-        return sorted(slots, key=lambda item: item.value)
+        return sorted(set(slots), key=lambda item: item.value)
 
     def _inferred_bases(self, context=None):
         # Similar with .ancestors, but the difference is when one base is inferred,
@@ -2761,7 +2890,7 @@ class ClassDef(mixins.FilterStmtsMixin, LocalsDictNodeNG, node_classes.Statement
                 bases_mro.append(ancestors)
 
         unmerged_mro = [[self]] + bases_mro + [inferred_bases]
-        _verify_duplicates_mro(unmerged_mro, self, context)
+        unmerged_mro = list(clean_duplicates_mro(unmerged_mro, self, context))
         return _c3_merge(unmerged_mro, self, context)
 
     def mro(self, context=None) -> List["ClassDef"]:
@@ -2774,7 +2903,7 @@ class ClassDef(mixins.FilterStmtsMixin, LocalsDictNodeNG, node_classes.Statement
         """
         return self._compute_mro(context=context)
 
-    def bool_value(self):
+    def bool_value(self, context=None):
         """Determine the boolean value of this node.
 
         :returns: The boolean value of this node.
